use crate::lib::ai::AIGoal;
use crate::lib::entity::{Ant, Egg, Entities, Food, Queen};
use crate::lib::point::Point;
use crate::lib::screen::{BoardCommand, Screen};
use std::collections::{HashMap, HashSet};

#[derive(Clone)]
pub struct World {
    pub cleared: HashMap<Point, Pheremone>,
    pub food: HashSet<Point>,
}

#[derive(Clone, PartialEq, PartialOrd, Debug)]
pub struct Pheremone {
    pub home: u32,
    pub food: u32,
}

impl Pheremone {
    pub fn new() -> Pheremone {
        Pheremone { home: 0, food: 0 }
    }

    pub fn decay(&mut self) {
        if self.home > 0 {
            self.home -= 1;
        }
        if self.food > 0 {
            self.food -= 1;
        }
    }
}

impl World {
    pub fn new() -> World {
        World {
            cleared: HashMap::new(),
            food: HashSet::new(),
        }
    }

    pub fn clear(&mut self, pos: Point) {
        self.cleared.insert(pos, Pheremone::new());
    }

    pub fn create_chamber(&mut self, center: Point, radius: i32) {
        let cx = center.0;
        let cy = center.1;

        for i in cx - radius..cx + radius {
            for j in cy - radius..cy + radius {
                self.clear(Point(i, j));
            }
        }
    }

    pub fn is_cleared(&self, pos: &Point) -> bool {
        self.cleared.contains_key(pos)
    }

    pub fn drop_pheremone(&mut self, pos: &Point, state: &AIGoal) {
        let op = self.cleared.get_mut(&pos);

        let ph = match op {
            Some(p) => p,
            None => {
                self.cleared.insert(*pos, Pheremone::new());
                self.cleared.get_mut(pos).unwrap()
            }
        };

        match state {
            AIGoal::Seek => ph.home += 1,
            AIGoal::Return => ph.food += 1,
        }
    }
}

pub fn render(e: &Entities, w: &World, b: &Screen) {
    for c in w.cleared.keys() {
        b.render(c, "x");
    }

    for a in e.data.values() {
        a.render(b);
    }
}

pub fn simulate(e: &mut Entities, w: &mut World, b: &mut Screen, step: u32) {
    let plan_cmds: Vec<BoardCommand> = e.data.values_mut().map(|a| a.plan(b, w)).collect();

    let mut cmds: Vec<BoardCommand> = e.data.values_mut().map(|a| a.step(b, w)).collect();

    cmds.extend(plan_cmds);

    for cmd in cmds {
        match cmd {
            BoardCommand::SpawnAnt => {
                let ant = Ant::new(b.center.0, b.center.1);
                e.add_entity(&ant);
            }
            BoardCommand::Dig(pos) => {
                w.clear(pos);
            }
            BoardCommand::LayEgg(pos, id) => {
                let egg = Egg::new(pos.0, pos.1, id);
                e.add_entity(&egg);
            }
            BoardCommand::Hatch(egg_id, queen_id) => {
                let egg = e.data.remove(&egg_id);
                let pos = egg.unwrap().get_position();

                let q = e.data.get_mut(&queen_id).unwrap();
                let queen: &mut Queen = q.downcast_mut::<Queen>().unwrap();
                queen.egg_count -= 1;
                let ant = Ant::new(pos.0, pos.1);
                e.add_entity(&ant);
            }
            BoardCommand::SpawnFood(pos) => {
                let food = Food::new(pos.0, pos.1);
                e.add_entity(&food);
                w.food.insert(pos);
            }
            BoardCommand::Noop => {}
        }
    }

    for n in b.center.get_neighbors() {
        w.drop_pheremone(&n, &AIGoal::Return);
    }
    w.drop_pheremone(&b.center, &AIGoal::Return);

    // decay all pheremones by some amount
    if step % 60 == 0 {
        for ph in w.cleared.values_mut() {
            ph.decay();
        }
    }
}
